@@ -11,7 +11,7 @@ import tempfile
 from fsspec import AbstractFileSystem
 from fsspec.compression import compr
 from fsspec.core import get_compression
-from fsspec.utils import stringify_path
+from fsspec.utils import isfilelike, stringify_path
 
 
 class LocalFileSystem(AbstractFileSystem):
@@ -19,7 +19,7 @@ class LocalFileSystem(AbstractFileSystem):
 
     Parameters
     ----------
-    auto_mkdirs: bool
+    auto_mkdir: bool
         Whether, when opening a file, the directory containing it should
         be created (if it doesn't already exist). This is assumed by pyarrow
         code.
@@ -118,10 +118,14 @@ class LocalFileSystem(AbstractFileSystem):
         elif self.isdir(path1):
             self.mkdirs(path2, exist_ok=True)
         else:
-            raise FileNotFoundError
+            raise FileNotFoundError(path1)
 
     def get_file(self, path1, path2, callback=None, **kwargs):
-        return self.cp_file(path1, path2, **kwargs)
+        if isfilelike(path2):
+            with open(path1, "rb") as f:
+                shutil.copyfileobj(f, path2)
+        else:
+            return self.cp_file(path1, path2, **kwargs)
 
     def put_file(self, path1, path2, callback=None, **kwargs):
         return self.cp_file(path1, path2, **kwargs)
@@ -154,7 +158,7 @@ class LocalFileSystem(AbstractFileSystem):
             self.makedirs(self._parent(path), exist_ok=True)
         return LocalFileOpener(path, mode, fs=self, **kwargs)
 
-    def touch(self, path, **kwargs):
+    def touch(self, path, truncate=True, **kwargs):
         path = self._strip_protocol(path)
         if self.auto_mkdir:
             self.makedirs(self._parent(path), exist_ok=True)
@@ -162,6 +166,8 @@ class LocalFileSystem(AbstractFileSystem):
             os.utime(path, None)
         else:
             open(path, "a").close()
+        if truncate:
+            os.truncate(path, 0)
 
     def created(self, path):
         info = self.info(path=path)
@@ -184,7 +190,7 @@ class LocalFileSystem(AbstractFileSystem):
         path = stringify_path(path)
         if path.startswith("file://"):
             path = path[7:]
-        return make_path_posix(path).rstrip("/")
+        return make_path_posix(path).rstrip("/") or cls.root_marker
 
     def _isfilestore(self):
         # Inheriting from DaskFileSystem makes this False (S3, etc. were)
@@ -192,6 +198,10 @@ class LocalFileSystem(AbstractFileSystem):
         # See https://github.com/dask/dask/issues/5526
         return True
 
+    def chmod(self, path, mode):
+        path = stringify_path(path)
+        return os.chmod(path, mode)
+
 
 def make_path_posix(path, sep=os.sep):
     """Make path generic"""
@@ -346,7 +356,7 @@ class LocalFileOpener(io.IOBase):
 
     def __enter__(self):
         self._incontext = True
-        return self.f.__enter__()
+        return self
 
     def __exit__(self, exc_type, exc_value, traceback):
         self._incontext = False
